﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace TestProject
{
    /// <summary>
    /// 색상 그리드
    /// </summary>
    public class ColorGrid : Control
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 선택 색상 변경시 - SelectedColorChanged

        /// <summary>
        /// 선택 색상 변경시
        /// </summary>
        public event EventHandler SelectedColorChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 행 수
        /// </summary>
        private const int ROW_COUNT = 5;

        /// <summary>
        /// 컬럼 수
        /// </summary>
        private const int COLUMN_COUNT = 8;

        /// <summary>
        /// 색상 문자열 배열
        /// </summary>
        private string[,] colorStringArray = new string[ROW_COUNT, COLUMN_COUNT]
        {
            { "Black"    , "Brown"    , "DarkGreen"  , "MidnightBlue", "Navy"     , "DarkBlue"    , "Indigo"   , "DimGray"   },
            { "DarkRed"  , "OrangeRed", "Olive"      , "Green"       , "Teal"     , "Blue"        , "SlateGray", "Gray"      },
            { "Red"      , "Orange"   , "YellowGreen", "SeaGreen"    , "Aqua"     , "LightBlue"   , "Violet"   , "DarkGray"  },
            { "Pink"     , "Gold"     , "Yellow"     , "Lime"        , "Turquoise", "SkyBlue"     , "Plum"     , "LightGray" },
            { "LightPink", "Tan"      , "LightYellow", "LightGreen"  , "LightCyan", "LightSkyBlue", "Lavender" , "White"     }
        };

        /// <summary>
        /// 색상 셀 배열
        /// </summary>
        private ColorCell[,] colorCellArray = new ColorCell[ROW_COUNT, COLUMN_COUNT];

        /// <summary>
        /// 선택 색상 셀
        /// </summary>
        private ColorCell selectedColorCell;

        /// <summary>
        /// 하이라이트 색상 셀
        /// </summary>
        private ColorCell highlightedColorCell;

        /// <summary>
        /// 테두리
        /// </summary>
        private Border border;

        /// <summary>
        /// 유니폼 그리드
        /// </summary>
        private UniformGrid uniformGrid;

        /// <summary>
        /// 선택 색상
        /// </summary>
        private Color selectedColor = Colors.Black;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 선택 색상 - SelectedColor

        /// <summary>
        /// 선택 색상
        /// </summary>
        public Color SelectedColor
        {
            get
            {
                return this.selectedColor;
            }
        }

        #endregion
        #region 비주얼 자식 수 - VisualChildrenCount

        /// <summary>
        /// 비주얼 자식 수
        /// </summary>
        protected override int VisualChildrenCount
        {
            get
            {
                return 1;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - ColorGrid()

        /// <summary>
        /// 생성자
        /// </summary>
        public ColorGrid()
        {
            this.border = new Border();

            this.border.BorderBrush     = SystemColors.ControlDarkDarkBrush;
            this.border.BorderThickness = new Thickness(1);


            AddVisualChild(this.border);

            AddLogicalChild(this.border);


            this.uniformGrid = new UniformGrid();

            this.uniformGrid.Background = SystemColors.WindowBrush;
            this.uniformGrid.Columns    = COLUMN_COUNT;

            this.border.Child = this.uniformGrid;

            for(int y = 0; y < ROW_COUNT; y++)
            {
                for(int x = 0; x < COLUMN_COUNT; x++)
                {
                    Color color = (Color)typeof(Colors).GetProperty(colorStringArray[y, x]).GetValue(null, null);

                    this.colorCellArray[y, x] = new ColorCell(color);

                    this.uniformGrid.Children.Add(this.colorCellArray[y, x]);

                    if(color == SelectedColor)
                    {
                        this.selectedColorCell = this.colorCellArray[y, x];

                        this.colorCellArray[y, x].IsSelected = true;
                    }

                    ToolTip toolTip = new ToolTip();

                    toolTip.Content = colorStringArray[y, x];

                    this.colorCellArray[y, x].ToolTip = toolTip;
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 비주얼 자식 구하기 - GetVisualChild(index)

        /// <summary>
        /// 비주얼 자식 구하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns></returns>
        protected override Visual GetVisualChild(int index)
        {
            if(index > 0)
            {
                throw new ArgumentOutOfRangeException("index");
            }

            return this.border;
        }

        #endregion
        #region 측정하기 (오버라이딩) - MeasureOverride(availableSize)

        /// <summary>
        /// 측정하기 (오버라이딩)
        /// </summary>
        /// <param name="availableSize">이용 가능한 크기</param>
        /// <returns>측정 크기</returns>
        protected override Size MeasureOverride(Size availableSize)
        {
            this.border.Measure(availableSize);

            return this.border.DesiredSize;
        }

        #endregion
        #region 배열하기 (오버라이딩) - ArrangeOverride(finalSize)

        /// <summary>
        /// 배열하기 (오버라이딩)
        /// </summary>
        /// <param name="finalSize">최종 크기</param>
        /// <returns>최종 크기</returns>
        protected override Size ArrangeOverride(Size finalSize)
        {
            this.border.Arrange(new Rect(new Point(0, 0), finalSize));

            return finalSize;
        }

        #endregion

        #region 마우스 ENTER 처리하기 - OnMouseEnter(e)

        /// <summary>
        /// 마우스 ENTER 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseEnter(MouseEventArgs e)
        {
            base.OnMouseEnter(e);

            if(this.highlightedColorCell != null)
            {
                this.highlightedColorCell.IsHighlighted = false;

                this.highlightedColorCell = null;
            }
        }

        #endregion
        #region 마우스 이동시 처리하기 - OnMouseMove(e)

        /// <summary>
        /// 마우스 이동시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            ColorCell colorCell = e.Source as ColorCell;

            if(colorCell != null)
            {
                if(this.highlightedColorCell != null)
                {
                    this.highlightedColorCell.IsHighlighted = false;
                }

                this.highlightedColorCell = colorCell;

                this.highlightedColorCell.IsHighlighted = true;
            }
        }

        #endregion
        #region 마우스 DOWN 처리하기 - OnMouseDown(e)

        /// <summary>
        /// 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseDown(MouseButtonEventArgs e)
        {
            base.OnMouseDown(e);

            ColorCell colorCell = e.Source as ColorCell;

            if(colorCell != null)
            {
                if(this.selectedColorCell != null)
                {
                    this.selectedColorCell.IsSelected = false;
                }

                this.selectedColorCell = colorCell;

                this.selectedColorCell.IsSelected = true;
            }

            Focus();
        }

        #endregion
        #region 마우스 UP 처리하기 - OnMouseUp(e)

        /// <summary>
        /// 마우스 UP 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseUp(MouseButtonEventArgs e)
        {
            base.OnMouseUp(e);

            ColorCell colorCell = e.Source as ColorCell;

            if(colorCell != null)
            {
                if(this.selectedColorCell != null)
                {
                    this.selectedColorCell.IsSelected = false;
                }

                this.selectedColorCell = colorCell;

                this.selectedColorCell.IsSelected = true;

                this.selectedColor = (this.selectedColorCell.Brush as SolidColorBrush).Color;

                OnSelectedColorChanged(EventArgs.Empty);
            }
        }

        #endregion
        #region 마우스 LEAVE 처리하기 - OnMouseLeave(e)

        /// <summary>
        /// 마우스 LEAVE 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseLeave(MouseEventArgs e)
        {
            base.OnMouseLeave(e);

            if(this.highlightedColorCell != null)
            {
                this.highlightedColorCell.IsHighlighted = false;

                this.highlightedColorCell = null;
            }
        }

        #endregion

        #region 키보드 포커스 획득시 처리하기 - OnGotKeyboardFocus(e)

        /// <summary>
        /// 키보드 포커스 획득시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnGotKeyboardFocus(KeyboardFocusChangedEventArgs e)
        {
            base.OnGotKeyboardFocus(e);

            if(this.highlightedColorCell == null)
            {
                if(this.selectedColorCell != null)
                {
                    this.highlightedColorCell = this.selectedColorCell;
                }
                else
                {
                    this.highlightedColorCell = this.colorCellArray[0, 0];
                }

                this.highlightedColorCell.IsHighlighted = true;
            }
        }

        #endregion
        #region 키 DOWN 처리하기 - OnKeyDown(e)

        /// <summary>
        /// 키 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);

            int index = this.uniformGrid.Children.IndexOf(this.highlightedColorCell);
            int y     = index / COLUMN_COUNT;
            int x     = index % COLUMN_COUNT;

            switch(e.Key)
            {
                case Key.Home :

                    y = 0;
                    x = 0;

                    break;

                case Key.End :

                    y = ROW_COUNT    - 1;
                    x = COLUMN_COUNT + 1;

                    break;

                case Key.Down :

                    if((y = (y + 1) % ROW_COUNT) == 0)
                    {
                        x++;
                    }

                    break;

                case Key.Up :

                    if((y = (y + ROW_COUNT - 1) % ROW_COUNT) == ROW_COUNT - 1)
                    {
                        x--;
                    }

                    break;

                case Key.Right :

                    if((x = (x + 1) % COLUMN_COUNT) == 0)
                    {
                        y++;
                    }

                    break;

                case Key.Left :

                    if((x = (x + COLUMN_COUNT - 1) % COLUMN_COUNT) == COLUMN_COUNT - 1)
                    {
                        y--;
                    }

                    break;

                case Key.Enter :
                case Key.Space :

                    if(this.selectedColorCell != null)
                    {
                        this.selectedColorCell.IsSelected = false;
                    }

                    this.selectedColorCell = this.highlightedColorCell;

                    this.selectedColorCell.IsSelected = true;

                    this.selectedColor = (this.selectedColorCell.Brush as SolidColorBrush).Color;

                    OnSelectedColorChanged(EventArgs.Empty);

                    break;

                default :

                    return;
            }

            if(x >= COLUMN_COUNT || y >= ROW_COUNT)
            {
                MoveFocus(new TraversalRequest(FocusNavigationDirection.Next));
            }
            else if(x < 0 || y < 0)
            {
                MoveFocus(new TraversalRequest(FocusNavigationDirection.Previous));
            }
            else
            {
                this.highlightedColorCell.IsHighlighted = false;

                this.highlightedColorCell = this.colorCellArray[y, x];

                this.highlightedColorCell.IsHighlighted = true;
            }

            e.Handled = true;
        }

        #endregion
        #region 키보드 포커스 상실시 처리하기 - OnLostKeyboardFocus(e)

        /// <summary>
        /// 키보드 포커스 상실시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnLostKeyboardFocus(KeyboardFocusChangedEventArgs e)
        {
            base.OnGotKeyboardFocus(e);

            if(this.highlightedColorCell != null)
            {
                this.highlightedColorCell.IsHighlighted = false;

                this.highlightedColorCell = null;
            }
        }

        #endregion

        #region 선택 색상 변경시 처리하기 - OnSelectedColorChanged(e)

        /// <summary>
        /// 선택 색상 변경시 처리하기
        /// </summary>
        /// <param name="e">이벤트 발생자</param>
        protected virtual void OnSelectedColorChanged(EventArgs e)
        {
            if(SelectedColorChanged != null)
            {
                SelectedColorChanged(this, e);
            }
        }

        #endregion
    }
}